//  RW Dolldrop
// ==============
// External File Transfers Library
// -Part of the RW Foundations suite of libraries

// Filename: ./include/dollhouse/net/dolldrop.hpp
// -Declarations of the baseline RW Dolldrop sub-library

#pragma once

///////
// Dolldrop/Dollnet is meant for network-enabled RW projects only
#if __has_include(<curl/curl.h>)

#include "incls_all.hpp"

#include <curl/curl.h>

#include "dollnet_util.hpp"

//////////////////
//  'Net Work :
//

// DESIGN:
// -As a simple, basic control mechanism for starters we'll use some
// protected/private classes and a few functions from here within Dolldrop.
// Later, we can add more access restrictions and some further data validations;
// improving things bit-by-bit as we go along, using just a few of the basic &
// straightforward language features readily available to us from the standard.

// DESIGN:
// - add work-stealing queues of the Dolldrops/Mailboxes ?
// - use S.Parent's libdispatch-based async
// - this is actually a general need, so add it to the RW Foundations instead

// DESIGN:
// void push_drop() {}
// void yank_drop() {}

// DESIGN: TODO:
// return do_atomiconn(
//   lockout();
//     connect();
//       do_req(uri);
//     disconnect();
//   unlockout();
// );

// DESIGN: during this prototyping phase, code-envy is acceptable. however, for
// the production code abstract common behaviours into modular functions.
// -DRY principle

namespace rw::dollhouse::net {

//------------------------------------------------------------------------------
/** The base class to manage dolldrops
 */
class Dolldrop_b : public RW_slave {
 public:
  Dolldrop_b(Hip const& tag, std::string const& handle);  // tag-parm'd ctor
                                                          //
  ~Dolldrop_b();                                          // to cleanup curl

 protected:
  /** Use @c curl-multi to pull requested headers in parallel
   * @param[in] tag Master's HIP
   * @param[in] pull_list the source pull list to check
   * @return the now-populated string container of pulled header's data
   */
  auto pull_hdrs_multi_g(Hip const& tag, Pull_list const& pull_list)
      -> std::vector<std::string>;

  /** Use @c curl-multi to pull requested bodies in parallel
   * @param[in] tag Master's HIP
   * @param[in] pull_list the source pull list to check
   */
  void pull_bodies_multi_g(Hip const& tag, Pull_list const& pull_list);

  /** Check and pull all headers that need updating, returning the updates list
   * @param[in] tag Master's HIP
   * @param[in] pull_list the source pull list to check
   * @return @c Pull_list container of @c Drop_detail 's needing updates
   */
  auto chk_hdrs_g(Hip const&                      tag,
                  Pull_list const&                pull_list,
                  std::vector<std::string> const& hdrs_dat) -> Pull_list;

  /** Ensure the dolldrops directory is in place to begin with
   * @return true on successful creation
   */
  auto mk_dropdir() -> bool;

  /** Write a general usage notes text file to dolldrops directory
   */
  void wrt_tailme();

  /** Execute a parallel-pull of the dolldrop items named in the pull listing
   * contained within @a list_path
   *
   * @param list_path
   * @note: wrapper for same member function in the private @c Dollnet field
   */
  void curl_pull_list_b(fs::path const& list_path);

  /** Signals a catalog pull request
   */
  bool is_catalog_{false};

 private:
  Dolldrop_b() = delete;  // no default ctor

  /** Initialize ./dolldrops/ directory
   */
  void preflight();

  //----------------------------------------------------------------------------
  /** The private member class to manage networking
   */
  class Dollnet : public RW_slave {
   public:
    Dollnet(Hip const& tag, std::string const& handle);  // tag-parm'd ctor

    /** Execute a parallel-pull of the dolldrop items named in the pull listing
     * contained within @a list_path
     *
     * This occurs by forking a process & running the @c curl program from a
     * shell to perform the downloads. This runs in parallel, and offers some
     * error-correction capabilities as well.
     * @param list_path
     * @note: simply shells out to call the @c curl program direct on the
     * terminal
     */
    void curl_pull_list(fs::path const& list_path);

   private:
    Dollnet() = delete;  // no default ctor

  } dollnet_;

};  // class Dolldrop_b

//------------------------------------------------------------------------------
/** The derived class to manage dolldrops
 */
class Dolldrop : public Dolldrop_b {
 public:
  // NOTE: no ctor; class interface only

  /** Attempts to pull, then check against locals, the headers of URIs contained
   * in @a pull_list
   *
   * @param[in] pull_list the source pull list to check
   * @return @c Pull_list container of @c Drop_detail 's needing updates
   * @note: simply marshals pull_bodies_multi_g(), chk_hdrs_g()
   */
  auto chk_hdrs(Pull_list const& pull_list) -> Pull_list;

  /** Assign to the *catalog pull request* signal
   * @param[in] flag
   * @note: controls state of protected member field in @c Dolldrop_b
   */
  void do_catalog(bool const flag);

  /** Attempts an external network connection with, and download transfer from,
   * @a uri. If successful, transfer's body stream for @a uri is saved out to
   * disk file (generic 'anons_file' form)
   *
   * @param[in] uri the URI to attempt a download connection with
   */
  void pull_anon_body(std::string const& uri);

  /** Execute a parallel-pull of the dolldrop items named in the pull listing
   * contained within @a list_path
   *
   * @param list_path
   */
  void curl_pull_list(fs::path const& list_path);

 private:
  Dolldrop() = delete;  // no default ctor

};  // class Dolldrop

}  // namespace rw::dollhouse::net

#endif  // __has_include(<curl/curl.h>)

//    --=^=--

// "______"

// rw_dollhouse_lib.v221022
// Copyright (2022)
// License MIT (Expat) https://opensource.org/licenses/MIT
