//  RW Dolldrop
// ==============
// External File Transfers Library
// -Part of the RW Foundations suite of libraries

// Filename: ./src/dollhouse/net/dolldrop.cpp
// -Definitions of the baseline RW Dolldrop sub-library

///////
// Dolldrop/Dollnet is meant for network-enabled RW projects only
#if __has_include(<curl/curl.h>)

#include "incls_all.hpp"

#include <curl/multi.h>

#include "dolldrop_util.cpp"

namespace rw::dollhouse::net {

//------------------------------------------------------------------------------
Dolldrop_b::Dolldrop_b(Hip const& tag, std::string const& handle)
    : RW_slave{{Type::Bumpmaster, Type::Curator}, handle, Type::Dolldrop},
      dollnet_{t(), "dolldrop_dollnet"} {
  stamp("Dolldrop_b()");

  bind_master(tag);

  preflight();
}

//------------------------------------------------------------------------------
Dolldrop_b::~Dolldrop_b() {
  curl_global_cleanup();
}

//------------------------------------------------------------------------------
/** Append a @c curl -specified memory block to a target string
 *
 * Writes an array of @a num_elems count of elements, each one @a elem_sz in
 * size, from the block of memory pointed to by @a addr, out to the current
 * position in the @a o_strm stream (the end of a @c std::string, in this case).
 *
 * @param[in] addr pointer to memory array of elements to be written out
 * @param[in] elem_sz size in bytes of each element
 * @param[in] num_elems number of elements
 * @param[out] o_strm pointer to output stream
 * @return number of bytes written
 * @note: this is intended as a callback function for a curl handle's use
 */
auto wrt_str_cb(Address const      addr,
                size_t const       elem_sz,
                size_t const       num_elems,
                std::string* const o_strm) -> size_t {  //

  auto const mem_blk_sz{elem_sz * num_elems};

  try {
    o_strm->append(static_cast<const char*>(addr), mem_blk_sz);

  } catch (exception const&) {
    // -"If an exception is thrown for any reason, this function has no effect
    //  (strong exception guarantee). (since C++11)"
    return 0;
  }

  return mem_blk_sz;

  // https://en.cppreference.com/w/cpp/string/basic_string/append
}

//------------------------------------------------------------------------------
/** Assign curl easy options for a new @c curl handle, adding it into a @c
 * curl-multi handle
 *
 * @param[out] multi_hdl the multi hdl that receives the new curl easy hdl
 * @param[in] uri the associated uri for this easy hdl
 * @param[in] hdr the target string to receive header data
 */
void cfg_hdr_hdl(CURLM* const       multi_hdl,
                 std::string const& uri,
                 std::string const& hdr) {
  auto const curl_hdl = curl_easy_init();

  curl_easy_setopt(curl_hdl, CURLOPT_WRITEFUNCTION, wrt_str_cb);

  curl_easy_setopt(curl_hdl, CURLOPT_NOBODY, 1L);
  curl_easy_setopt(curl_hdl, CURLOPT_HEADERDATA, &hdr);

  curl_easy_setopt(curl_hdl, CURLOPT_USERAGENT, "Bumpmaster_v221022");

  curl_easy_setopt(curl_hdl, CURLOPT_URL, uri.c_str());
  curl_easy_setopt(curl_hdl, CURLOPT_FOLLOWLOCATION, 1L);

  curl_multi_add_handle(multi_hdl, curl_hdl);
}

//------------------------------------------------------------------------------
auto Dolldrop_b::pull_hdrs_multi_g(Hip const& tag, Pull_list const& pull_list)
    -> std::vector<std::string> {
  stamp(tag, "pull_hdrs_multi_g()");

  auto const cnt = pull_list.size();

  std::vector<std::string> hdrs;
  hdrs.reserve(cnt);

  if (tag == master()) {  // function is meant only for slave's Master to use

    cout << "\n checking (" << cnt << ") item's headers..." << flush;

    auto const multi_hdl = curl_multi_init();

    // load up the curl-multi handle, w/ each individual curl handle per header
    for (auto const& [uri, fs_path, idx] : pull_list) {  //

      hdrs.emplace_back("");  // start w/ an empty
      auto const& hdr = hdrs.back();

      cfg_hdr_hdl(multi_hdl, uri, hdr);
    }

    ///

    auto           running_hdls{0};
    constexpr auto timeout_ms{20 * 1000};  // 20s (increase if needed, Anon)

    // perform the curl-multi operation, waiting for stragglers to complete
    //
    curl_multi_perform(multi_hdl, &running_hdls);
    do {
      auto ret{0};
      auto res_code = curl_multi_wait(multi_hdl, nullptr, 0, timeout_ms, &ret);

      if (res_code != CURLM_OK) {
        ;  // TBD...
      }

      curl_multi_perform(multi_hdl, &running_hdls);

    } while (running_hdls);

    ///

    // cleanups-only (messages ignored for now; TBD...)
    // -DESIGN: abstract this out if no further use comes up here
    //
    auto     msgs_in_queue{0};
    CURLMsg* msg = nullptr;
    //
    while ((msg = curl_multi_info_read(multi_hdl, &msgs_in_queue))) {
      auto curl_hdl = msg->easy_handle;
      curl_multi_remove_handle(multi_hdl, curl_hdl);
      curl_easy_cleanup(curl_hdl);
    }
    //
    curl_multi_cleanup(multi_hdl);

  } else
    only_master_can(master(), tag, t(), "Dolldrop_b::pull_hdrs_multi_g()");

  return hdrs;
}

//------------------------------------------------------------------------------
/** Append a @c curl -specified memory block to a target file stream
 *
 * Writes an array of @a num_elems count of elements, each one @a elem_sz in
 * size, from the block of memory pointed to by @a addr, out to the current
 * position in the @a o_strm stream (the end of @c std::ofstream, in this case).
 *
 * @param[in] addr pointer to memory array of elements to be written out
 * @param[in] elem_sz size in bytes of each element
 * @param[in] num_elems number of elements
 * @param[out] o_strm pointer to output stream
 * @return number of bytes written
 * @note: this is intended as a callback function for a curl handle's use
 */
auto wrt_strm_cb(Address const        addr,
                 size_t const         elem_sz,
                 size_t const         num_elems,
                 std::ofstream* const o_strm) -> size_t {  //
  auto const mem_blk_sz{elem_sz * num_elems};

  try {
    o_strm->write(static_cast<const char*>(addr), mem_blk_sz);

    // DESIGN: TODO:
    // -read 'badbit' & throw if true

  } catch (exception const&) {
    // DESIGN: TODO:
    // get/(re)set stream posn

    return 0;
  }

  return mem_blk_sz;
}

//------------------------------------------------------------------------------
/** Assign curl easy options for a new @c curl handle, adding it into a @c
 * curl-multi handle
 *
 * @param[out] multi_hdl the multi hdl that receives the new curl easy hdl
 * @param[in] uri the associated uri for this easy hdl
 * @param[in] body the target filestream to receive body data
 */
void cfg_body_hdl(CURLM* const         multi_hdl,
                  std::string const&   uri,
                  std::ofstream const& body) {
  auto const curl_hdl = curl_easy_init();

  curl_easy_setopt(curl_hdl, CURLOPT_WRITEFUNCTION, wrt_strm_cb);
  curl_easy_setopt(curl_hdl, CURLOPT_WRITEDATA, &body);

  // DESIGN: will necessitate some form of screen mgmt w/ multi-bodies usage
  // -perhaps consider the 'blessed' TUI libs for some inspiration?
  //   https://github.com/chjj/blessed
  //   https://github.com/yaronn/blessed-contrib
  // -JS-based, so obvs non-starters. OTOH, they far surpass ncurses' look IMO
  //
  curl_easy_setopt(curl_hdl, CURLOPT_XFERINFOFUNCTION, prt_prog);
  curl_easy_setopt(curl_hdl, CURLOPT_NOPROGRESS, 0L);
  //

  curl_easy_setopt(curl_hdl, CURLOPT_USERAGENT, "Bumpmaster_v221022");

  curl_easy_setopt(curl_hdl, CURLOPT_URL, uri.c_str());
  curl_easy_setopt(curl_hdl, CURLOPT_FOLLOWLOCATION, 1L);

  curl_multi_add_handle(multi_hdl, curl_hdl);
}

//------------------------------------------------------------------------------
void Dolldrop_b::pull_bodies_multi_g(Hip const&       tag,
                                     Pull_list const& pull_list) {
  stamp(tag, "pull_bodies_multi_g()");

  if (tag == master()) {  // function is meant only for slave's Master to use

    vector<ofstream> bodies;
    bodies.reserve(pull_list.size());

    auto const multi_hdl = curl_multi_init();

    // load up the curl-multi handle, w/ each individual curl handle per body
    for (auto const& [uri, fs_path, idx] : pull_list) {  //

      bodies.emplace_back(ofstream{fs_path});
      auto const& body = bodies.back();

      cfg_body_hdl(multi_hdl, uri, body);
    }

    ///

    auto           running_hdls{0};
    constexpr auto timeout_ms{20 * 1000};  // 20s (increase if needed, Anon)

    // DESIGN: will necessitate multiple timers w/ multi-bodies usage
    start = Call_clk::now();

    // perform the curl-multi operation, waiting for stragglers to complete
    //
    curl_multi_perform(multi_hdl, &running_hdls);
    do {
      auto ret{0};  // unused atm...
      auto res_code = curl_multi_wait(multi_hdl, nullptr, 0, timeout_ms, &ret);

      if (res_code != CURLM_OK) {
        ;  // TBD...
      }

      curl_multi_perform(multi_hdl, &running_hdls);

    } while (running_hdls);

    ///

    // cleanups-only (messages ignored for now; TBD...)
    // -DESIGN: abstract this out if no further use comes up here
    //
    auto     msgs_in_queue{0};
    CURLMsg* msg = nullptr;
    //
    while ((msg = curl_multi_info_read(multi_hdl, &msgs_in_queue))) {
      auto curl_hdl = msg->easy_handle;
      curl_multi_remove_handle(multi_hdl, curl_hdl);
      curl_easy_cleanup(curl_hdl);
    }
    //
    curl_multi_cleanup(multi_hdl);

    // _always_ close output file streams Anon :^)
    for (auto& body : bodies)
      body.close();

    puts("\n");

  } else
    only_master_can(master(), tag, t(), "Dolldrop_b::pull_bodies_multi_g()");
}

//------------------------------------------------------------------------------
auto Dolldrop_b::chk_hdrs_g(Hip const&                      tag,
                            Pull_list const&                pull_list,
                            std::vector<std::string> const& hdrs_dat)
    -> Pull_list {
  stamp(tag, "chk_hdrs_g()");

  Pull_list res;

  if (tag == master()) {  // function is meant only for slave's Master to use

    // check header strings contained in hdrs_dat, against the data in droplog
    //
    for (auto const& [uri, fs_path, idx] : pull_list) {
      stringstream hdr{hdrs_dat[idx]};  // an already-retrieved hdr for this uri

      // If header's DL failed for some reason, just attempt the curl DL
      // anyway
      if (is_empty(hdr)) {
        // DESIGN:
        // -any confirmation of valid file indicated by fs_path needed here?

        res.emplace_back(make_tuple(uri, fs_path, idx));
        continue;  // skip further proc'ng for this dolldrop's header
      }

      // check if the remote source (claims to be) modified since last
      // download
      // -DESIGN: 'name' now unused, drop from called func ret
      //
      const auto& [have_new_drop, name] = chk_droptimes(uri, hdr);

      if (have_new_drop) {  // yes, remote 'last-modified' header tag differs

        // check if the remote source's and local dolldrop's 'length' tags
        // match
        const auto size_match = chk_dropsizes(uri, hdr);

        // -if this is a catalog's pull check, just a newer remote file is
        // good enough (since it's size may stay identical, even w/ a
        // legitimate change)
        // -otherwise, also require that the size of the remote differs from
        // local
        //
        if (is_catalog_ || (! size_match))  // yes, add item to pull list
          res.emplace_back(make_tuple(uri, fs_path, idx));
      }
    }

  } else
    only_master_can(master(), tag, t(), "Dolldrop_b::chk_hdrs_g()");

  return res;
}

//------------------------------------------------------------------------------
auto Dolldrop_b::mk_dropdir() -> bool {
  stamp("mk_dropdir()");

  fs::create_directory(dropdir);
  return exists_fs(dropdir);
}

//------------------------------------------------------------------------------
void Dolldrop_b::wrt_tailme() {  //
  stamp("wrt_tailme()");

  // a raw string; the body of the text file
  auto const tailtext =
      R"(//  RW Dolldrop
// ==============
// External File Transfers Library
// -Part of the RW Foundations suite of libraries

Notes:

-You can use the tail command to watch DLs progressively fill your dolldrops
 E.G., (in a terminal from the project's base directory) :

    tail -f  ./dolldrops/anons_file_001

-You can even playback a stream-capp style dolldrop 'live' (while its capturing)
 E.G. :

    vlc  ./dolldrops/anons_file_001
       or
    mpv  ./dolldrops/anons_file_042.ogg

-Protip : If you want a file extension automatically appended to your dolldrops,
 put it on the second line of the file named :  ./dolldrops/tagfile
 E.G., for OGG dolldrops, a tagfile might be edited to begin with the contents :

    000
    .ogg

 which would (on repeated use of this demo) create a series of dolldrops named :

    ./dolldrops/anons_file_001.ogg
    ./dolldrops/anons_file_002.ogg
    ./dolldrops/anons_file_003.ogg

 &tc. (btw this can be changed anytime to get a different extension next round)

Cheers.

// rw_dollhouse_lib
// Copyright (2022)
// License MIT (Expat) https://opensource.org/licenses/MIT
)";

  ofstream ofs{dropdir / "tail_me.txt"};
  ofs << tailtext;
  ofs.close();
}

//------------------------------------------------------------------------------
void Dolldrop_b::curl_pull_list_b(fs::path const& list_path) {
  stamp("curl_pull_list_b()");

  dollnet_.curl_pull_list(list_path);
}

//------------------------------------------------------------------------------
void Dolldrop_b::preflight() {  //
  stamp("preflight()");

  curl_global_init(CURL_GLOBAL_ALL);

  // check if ./dolldrops/ directory exists
  if (! exists_fs(dropdir)) {
    try {
      mk_dropdir();
      wrt_tailme();

    } catch (fs::filesystem_error const& e) {
      cout << "  fs::filesystem_error : '" << e.what()
           << "' caught in Dolldrop_b::preflight()\n\n";

    } catch (std::bad_alloc const& e) {
      cout << "  std::bad_alloc : '" << e.what()
           << "' caught in Dolldrop_b::preflight()\n\n";
    }
  }
}

//------------------------------------------------------------------------------
auto Dolldrop::chk_hdrs(Pull_list const& pull_list) -> Pull_list {  //
  stamp("chk_hdrs()");

  auto const hdrs_dat = pull_hdrs_multi_g(master(), pull_list);

  return chk_hdrs_g(master(), pull_list, hdrs_dat);
}

//------------------------------------------------------------------------------
void Dolldrop::do_catalog(bool const flag) {
  stamp("do_catalog()");

  is_catalog_ = flag;
}

//------------------------------------------------------------------------------
void Dolldrop::pull_anon_body(std::string const& uri) {  //
  stamp("pull_anon_body()");

  putchar('\n');
  prt_tag(master(), "asked to pull a dolldrop");
  cout << "  -requested pull address :  " << uri << "\n\n";

  // add just a single dolldrop to the pull list in this case
  auto      dropname = bld_dropname();
  Pull_list pull_list{Drop_detail{uri, dropname, 0}};

  pull_bodies_multi_g(master(), pull_list);
}

//------------------------------------------------------------------------------
void Dolldrop::curl_pull_list(fs::path const& list_path) {
  stamp("curl_pull_list()");

  curl_pull_list_b(list_path);
}

}  // namespace rw::dollhouse::net

#endif  // __has_include(<curl/curl.h>)

//    --=^=--

// "______"

// rw_dolldrop_lib.v221022
// Copyright (2022)
// License MIT (Expat) https://opensource.org/licenses/MIT
